/*
  TFM-CUDA
  Copyright (C) 2008-2010  Mathieu Giraud & Jean-Stéphane Varré
  http://bioinfo.lifl.fr/TFM/TFM-CUDA/ 

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>

int conf_N = 0 ; // automatic N
int conf_print_plots = 0 ;
int conf_verbose = 0 ;
int conf_verbose_buck = 0 ;
int conf_last_combine = 1 ;
int conf_device = 0 ;

#include "pwm.h"

#include "time_tools.c"
#include "cuda_tools.c"
#include "matrix.c"
#include "device_tools.c"
#include "word_tools.c"


// Score buckets 

#include "histograms.c"
#include "score_buckets.c"

// Scan

#include "scan.c"



void usage(char *progname)
{
   printf("Usage: %s [options] matrix.pfm [matrix2.pfm ...] \n", progname);

   printf("\n  -S    Scan (N matrices)\n");
   printf("         -T #  set the score threshold\n");
   printf("         -P #  set the Pvalue exponent (default used 5)\n");
   printf("         -l #  random sequence length\n");
   printf("         -s #  sequence file\n");
   printf("         -H    display the hits\n");

   printf("\n  -B    BucketScoreDistribution %s\n",

#ifdef COMPARE
	  ": compare (EXPERIMENTAL) (2 matrices)"
#else
	  ": pvalue-to-score / score-to-pvalue (1 matrix)"
#endif
	  );


   printf("         -P #  set the Pvalue exponent (score-to-pvalue) (0 loops on all exponents)\n");
   printf("         -T #  set the score threshold (pvalue-to-score)\n");


   printf("\n         algorithm options\n");
   printf("         -N #  set the number of slices (0 for automatic)\n");
   printf("         -c    do not do the last combine (pvalue-to-score only)\n");
   printf("\n         CUDA options\n");
   printf("         -d #  select CUDA device\n");
   printf("\n         verbose options\n");
   printf("         -p    print plots for the score distributions\n");
   printf("         -v    verbose\n");


	      
}


int main(int argc, char **argv)
{
   // print_sizeofs();
	
   extern int optind ;
   extern char *optarg ;
   char c ;
   int conf_Pvalue = 5 ;
   float conf_threshold = 0.0 ;
   int conf_display_hits = 0 ;

   int conf_scan = 0 ;
   int conf_buck = 0 ;
   int conf_random_length = 50 * 1024 * 1024 ;
   const char* conf_seq_file = NULL ; 

   int conf_optpvalue = 0;
   int conf_optthreshold = 0;
   
   while ((c = getopt(argc, argv, "P:T:pvcN:SBl:s:Hd:")) != EOF)
      switch(c)
	 {
	 case 'H':
	    conf_display_hits = 1 ;
	    break ;

	 case 'N':
	    conf_N = atoi(optarg) ;
	    break ;

	 case 'l':
	    conf_random_length = atoi(optarg) ;
	    break ;

	 case 'T':
	    conf_optthreshold = 1;
	    conf_threshold = atof(optarg) ;
	    break ;

	 case 'P':
            conf_optpvalue = 1;
	    conf_Pvalue = atoi(optarg) ;
	    if (!conf_Pvalue)
	       conf_last_combine = 0 ;
	    break ;

	 case 'p':
	    conf_print_plots = 1 ;
	    break ;

	 case 's':
	    conf_seq_file = optarg ;
	    break ;

	 case 'v':
	    conf_verbose = 1 ;
	    conf_verbose_buck = 1 ;
	    break ;

	 case 'c':
	    conf_last_combine = 1 ;
	    break ;

	 case 'd':
            conf_device = atoi(optarg);
            break ;

	 case 'S':
	    conf_scan = 1 ;
	    break ;

	 case 'B':
	    conf_buck = 1 ;
	    conf_verbose_buck = 1 ;
	    break ;

	 default:
	    usage(argv[0]);
	    return 1 ;
	 }

   if (conf_scan + conf_buck == 0)
      {
	 printf("One command (-S / -B) is mandatory.\n");
	 usage(argv[0]);
	 exit(1);
      }

   if (conf_optpvalue + conf_optthreshold != 1)
      {
	 printf("You have to choose between setting the The Pvalue parameter (-P) or the threshold parameter (-T)\n");
	 usage(argv[0]);
	 exit(1);
      }

   if (conf_scan && (!conf_Pvalue))
      {
	 printf("The Pvalue parameter (-P) must be greater than 0 for the scan command\n");
	 usage(argv[0]);
	 exit(1);
      }

   printf("== prog: %s%s ", argv[0], PROG_FLAVOUR);
   printf(" -N %d -T %f -P %d \n", conf_N, conf_threshold, conf_Pvalue);
	 

   // read the matrices
   int nb_matrices = argc - optind ;

   if (!nb_matrices)
      {
	 printf("!! No matrix given : using default matrix %s \n", DFT_MATRIX);
	 nb_matrices = 1 ;		
      }

   float *matrices = (float *) malloc(nb_matrices * ALPH_SIZE * (MATRIX_MAX_LENGTH) * sizeof(float));
   int *lengths = (int *) malloc(nb_matrices * sizeof(int)) ;
   int offset = 0 ;

   int i = 0 ;
   while ((optind < argc) || (i == 0))
      {
	 if (optind == argc)
	    matrix_from_file(matrices, DFT_MATRIX, lengths);
	 else
	    matrix_from_file(matrices + offset, argv[optind++], lengths + i) ;
		
	 offset += ALPH_SIZE * lengths[i] ;
	 i++ ;

	 if (conf_buck)
	    {
	       if (conf_N * MAX_LENGTH_PER_SLICE < lengths[i])
		  {
		     printf("Matrix too large : please use a larger N, or increase MAX_LENGTH_PER_SLICE.\n");
		     exit (6);
		  }

	    }
      }
	
   printf(" <--  %d matrices, total length %d \n", nb_matrices, offset / ALPH_SIZE);

	    
#ifdef COMPARE
   if (! (lengths[0] == lengths[1]))
      {
	 printf("Prototype : the two matrices must have the same length.\n");
	 exit(6);
      }
#endif

   // float *matrix = random_matrix(length) ;


   int seq_length ;
   unsigned char *seq ;

   if (conf_scan)
      {

	 if (conf_seq_file)
	    seq = seq_from_file(conf_seq_file, &seq_length);
					
	 else
	    {
	       // random seq
	       seq_length = conf_random_length ;
	       seq = random_seq(seq_length + MATRIX_MAX_LENGTH);
	    }
      }

   if (optind<argc) // non atteint, TODO
      {
	 printf("Too many arguments.\n");
	 usage(argv[0]);
	 return 1 ;
      }


   // ----------------------- 

   time_start();  
   init_cuda(conf_device);

   if (conf_scan)
      {
	 // make the array of thresholds
	 float thresholds[nb_matrices];
	 
	 if (conf_optpvalue) 
	    {
	      printf("\n==== Compute score thresholds ====\n");
	       // compute the score threshold for each matrix
	       int offset = 0;
	       for (int i = 0; i < nb_matrices; i++)
		  {
		     pvalue2score_n(matrices + offset, lengths[i], conf_N, conf_Pvalue, conf_threshold, &conf_threshold);  
		     thresholds[i] = conf_threshold;
		     printf("==> Matrix %i, threshold = %f\n\n",i,conf_threshold);

		     offset += ALPH_SIZE * lengths[i];
		  }

	       time_stop();	
	    }
	 else 
	    {
	       // otherwise set the same threshold
	       for (int i = 0; i < nb_matrices; i++)
		  thresholds[i] = conf_threshold;
	    }
	 scan(matrices, nb_matrices, lengths, seq, seq_length, thresholds, conf_display_hits);
      }


   if (conf_buck)
      {
	 float score;
	 pvalue2score_n(matrices, 
#ifdef COMPARE
			matrices + lengths[0],
#endif
			lengths[0], conf_N, conf_Pvalue, conf_threshold, &score);  
      }


   printf("==\n");
   time_stop();	

}
