/*
  TFM-CUDA
  Copyright (C) 2008-2010  Mathieu Giraud & Jean-Stéphane Varré
  http://bioinfo.lifl.fr/TFM/TFM-CUDA/ 

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


// depending on the use of the slice data structure, matrix is the matrix itself or the slice data structure
__device__ float pwm_score(const float* matrix, 
			   int* word, int length, float threshold, 
			   float* best_scores, float *worst_scores, int numMatrix)
{

   float score = 0.0 ;
   int *w = word;

#ifdef SLICES
   int nb_slices = length / SLICEWIDTH;
   if (length % SLICEWIDTH) nb_slices++;
   for (int slice = 0; slice < nb_slices; slice++) {
      // compute the offset of the score of the given word
      long intword = 0;
      for (int pos = SLICEWIDTH-1; pos >= 0; pos--) {
	 //PRINTF("%1d",*w);
	 intword = intword * ALPH_SIZE + (*(w+pos));
      }
      w += SLICEWIDTH;
      score += matrix[slice * SLICESIZE * NBMAT_PER_SLICE + SLICESIZE * numMatrix + intword];
      //PRINTF(" - [%1d %ld] %.2f -> %.2f\n", slice, intword, matrix[slice * SLICESIZE * NBMAT_PER_SLICE + SLICESIZE * numMatrix + intword], score);
   }
#else
   int t;
   for (t=0; t<length; t++) {
      score += matrix[ALPH_SIZE * t + (*w)] ;
      w++;
   }
#endif
  

   DEBUG printf_word(word, length);
   DEBUG PRINTF("  %5.8f \n", score);

   return (score) ;
}


__device__ void compute_best_worst_scores(float *matrix, int length, float *best_scores, float *worst_scores)
{
   // temps: +1%
   // to parallelize, operations dupliquees pour l'instant...

   int v = (length+1) * ALPH_SIZE - 1;
   int l;
   best_scores[length+1] = worst_scores[length+1] = 0 ;

   for (l=length; l>=0; l--)
      { 
	 float best = matrix[v] ;
	 float worst = matrix[v] ;
	
	 v-- ;

	 int a ;
	 for (a=1; a<ALPH_SIZE; a++)
	    {
	       if (matrix[v] > best)
		  best = matrix[v] ;
	    
	       if (matrix[v] < worst)
		  worst = matrix[v] ;
	
	       v-- ;
	    }

	 best_scores[l] = best_scores[l+1] + best ;
	 worst_scores[l] = worst_scores[l+1] + worst ;
	
	 // DEBUG PRINTF(" l:%d   best:%.4f  worst:%.4f  \n", l, best_scores[l], worst_scores[l]);
      }
 
#ifdef SLICES
   for (l=length; l>=0; l--)
      {
	 best_scores[2*l] = best_scores[l];
	 worst_scores[2*l] = worst_scores[l];
      }
#endif
}

  

// matrices_device  : the matrices arranged one behind the other
// matrices_lengths : length of each matrices
// nb_matrices      : the number of matrices for which we have to compute the score
// thresholds       : the score thresholds for each matrix
// results_mat      : the matrix number for the i-th result
// results_pos      : the position in the sequence for the i-th result
// results_scores   : the score of the matrix for the i-th result
// results_occ      : the number of occurrences computed by the current thread
__global__ void scan_pwm (float *matrices_device, int *matrices_lengths, int nb_matrices, int offset_num_matrix, 
			  float *thresholds, 
			  int *results_mat, int* results_pos, float* results_scores, int* results_occ,
			  unsigned char* seq_device, int seq_length
#ifdef SLICES
			  , float *device_tab = NULL
#endif
			  )
{

#ifdef SLICES
   // DEBUG
   DEBUG {
      PRINTF("Slice tab in scan_pwm\n");
   int s;
   for (s = 0; s < NBSLICES; s++) {
      PRINTF("slice %2d : [ ",s);
      int w;
      for (w = 0; w < SLICESIZE * NBMAT_PER_SLICE; w++) {
	 PRINTF("%.2f ",*(device_tab + s*SLICESIZE*NBMAT_PER_SLICE+w));
      }
      PRINTF("]\n");
   }
   }
#endif 

#ifdef WU_BEST
   // compute best/worst scores
   __shared__ float best_scores[MATRIX_MAX_LENGTH_SCAN] ;
   __shared__ float worst_scores[MATRIX_MAX_LENGTH_SCAN] ;
   compute_best_worst_scores(matrix_shared, length, best_scores, worst_scores);
#endif


   // Determine my id
   const unsigned int bx = blockIdx.x;
   const unsigned int tx = threadIdx.x;
   long id = bx * NB_THREADS_PER_BLOCK_SCAN + tx ;
   // PRINTF("Block %5d, thread %5d, id %6ld \n ", bx, tx, id);

   int *results_mat_out      = results_mat    + id * RES_PER_THREAD ;
   int *results_pos_out      = results_pos    + id * RES_PER_THREAD ;
   float *results_scores_out = results_scores + id * RES_PER_THREAD ;

   // Copy the seq
   // long slices_per_block = 1 ;
   long segments_per_block = 1 + seq_length / (NB_BLOCKS_SCAN * SEGMENT_LENGTH_SHARED)  ;
   __shared__ int seq_shared[SEGMENT_LENGTH_SHARED + MATRIX_MAX_LENGTH_SCAN] ;    


   //  int t; for (t=0; t<16; t++) // for benchmarks
   {

      int nbocc = 0 ;
      int *seq ;
      int segment = 0;

#ifdef SLICES
#ifdef SLICES_IN_SHARED
      __shared__ float shared_tab[NBSLICES * SLICESIZE * NBMAT_PER_SLICE];
      copy_float_device_shared(shared_tab, device_tab, NBSLICES * SLICESIZE * NBMAT_PER_SLICE);
#endif
#else // NO SLICES
#ifdef LATE_MATRIX_COPY  
      __shared__ float matrix[ALPH_SIZE * MATRIX_MAX_LENGTH_SCAN] ;
#else // NO LATE_MATRIX_COPY
      __shared__ float matrices_shared[ALPH_SIZE * MAX_TOTAL_LENGTH] ;
      copy_float_device_shared(matrices_shared, matrices_device, ALPH_SIZE * MAX_TOTAL_LENGTH);
#endif
#endif

      for (segment = 0; segment < segments_per_block; segment++)
	 {
	    int segment_offset = SEGMENT_LENGTH_SHARED * (bx*segments_per_block + segment) ;
	    copy_seq_device_shared(seq_shared, seq_device + segment_offset, SEGMENT_LENGTH_SHARED + MATRIX_MAX_LENGTH_SCAN);

#ifndef SLICES
#ifdef LATE_MATRIX_COPY
	    float *matrix_device = matrices_device; // iterator over the matrices
#else // NO LATE_MATRIX_COPY
	    float *matrix = matrices_shared; // will contain the matrix
#endif
#endif

	    int n;
	    for (n = 0; n < nb_matrices; n++) 
	       {

		  int length = matrices_lengths[n];

#ifndef SLICES
#ifdef LATE_MATRIX_COPY
		  //for (int i = 0; i < ALPH_SIZE * MATRIX_MAX_LENGTH_SCAN; i++) *(matrix + i) = 0.0;
		  //for (int i = 0; i < ALPH_SIZE * length; i++) *(matrix + i) = *(matrix_device + i);
		  //__syncthreads();
		  copy_float_device_shared(matrix, matrix_device, length * ALPH_SIZE);
#endif
		  // DEBUG
		  DEBUG {
		     PRINTF("Matrix %d of length %d\n",n,length);
		     for (int x = 0; x < ALPH_SIZE; x++) {
			for (int pos = 0; pos < MATRIX_MAX_LENGTH_SCAN; pos++) {
			   PRINTF("% 4.2f ",*(matrix + ALPH_SIZE * pos + x));
			}
			PRINTF("\n");
		     }
		     PRINTF("\n");
		  }
#endif

#ifdef THREADS_INTERLEAVED
		  // seq = seq_shared + tx;
		  for (seq = seq_shared + tx; seq < seq_shared + SEGMENT_LENGTH_SHARED; seq+= NB_THREADS_PER_BLOCK_SCAN)
#else
		     int SEQ_LENGTH_BY_THREAD = SEGMENT_LENGTH_SHARED / NB_THREADS_PER_BLOCK_SCAN ;
		  for (seq = seq_shared + tx * SEQ_LENGTH_BY_THREAD; 
		       seq < seq_shared + (tx+1) * SEQ_LENGTH_BY_THREAD; 
		       seq+= 1)
#endif
		     {

			if ((seq - seq_shared) + segment_offset + length >= seq_length)
			   break ;

			float score = pwm_score(
#ifdef SLICES
#ifdef SLICES_IN_SHARED
						shared_tab,
#else
						device_tab,
#endif
#else
						matrix,
#endif
						seq, length, thresholds[n], 
#ifdef WU_BEST
						best_scores, worst_scores
#else
						NULL, NULL
#endif
						, n);  
			

			int good = (score >= thresholds[n]) ;

			if (good)
			   {
			      // DEBUG
			      //for (int i = 0; i < length; i++) PRINTF("%d",seq[i]);
			      //PRINTF(" %d \n", (seq-seq_shared) + segment_offset);
   

			      results_pos_out[nbocc % RES_PER_THREAD] = (seq-seq_shared) + segment_offset ; 
			      results_mat_out[nbocc % RES_PER_THREAD] = n + offset_num_matrix;
#ifdef SCORES_OUT
			      results_scores_out[nbocc % RES_PER_THREAD] = score ; 
#endif

			      nbocc++ ;
			   }
		     } // for seq

#ifndef SLICES
#ifdef LATE_MATRIX_COPY
		  matrix_device += ALPH_SIZE * length ;
#else
		  matrix += ALPH_SIZE * length ;
#endif
#endif
		 
	       } // for matrices
	 } // for segment
      DEBUG PRINTF("  --> %5d occurences\n", nbocc);

      __syncthreads();

      results_occ[id] = nbocc ;
   }  

}


// 
void device_group_multiple_matrices(float *res, float *matrices, int *lengths, int nb_matrices) 
{

   // initializing res
   {
      float *roffset = res;
      for (int slice = 0; slice < NBSLICES; slice++) {
	 for (int i = 0; i < SLICESIZE * NBMAT_PER_SLICE; i++) {
	    (*roffset) = 0.0;
	    roffset++;
	 }
      }
   }
   
   // computing values of the slice tables
   {
      int l = 0;
      while (l < SLICEWIDTH) { // must be maximal slice length
	 long nbSpan = 1 << (2 * l); // number of successive cases with same score to be added
	 long nbStep = nbSpan * ALPH_SIZE;    // increasing step
	 //		PRINTF("span = %ld step %ld\n",nbSpan, nbStep);
	 for (int slice = 0; slice < NBSLICES; slice++) {
	    float *roffset = res + slice * SLICESIZE * (NBMAT_PER_SLICE);
	    if (l < SLICEWIDTH) {
	       for (int x = 0; x < ALPH_SIZE; x++) {
		  for (long step = 0; step < SLICESIZE; step += nbStep) {
		     for (long span = 0; span < nbSpan; span++) {
			float *matrix = matrices;
			for (int n = 0; n < nb_matrices; n++) {
			   if (SLICEWIDTH * slice + l < lengths[n]) {
			      *(roffset + SLICESIZE * n + x * nbSpan + step + span) += *(matrix + slice * (SLICEWIDTH * ALPH_SIZE) + l * ALPH_SIZE + x);
			   } 
			   matrix += ALPH_SIZE * lengths[n];
			}
		     }
		  }
	       }
	    }
	 }
	 l++;
      }   
   }

      
}

void scan(float *matrices, int nb_matrices, int *lengths,  unsigned char *seq, int seq_length, float *thresholds, int display_hits)
{
   printf("== Seq (%d) : ", seq_length) ;
   printf_seq(seq, 50);
   printf("...\n");

   // allocate device memory for matrices and copy them
   float *device_matrices; 
   cudaMalloc((void**) &device_matrices, nb_matrices * ALPH_SIZE * MATRIX_MAX_LENGTH * sizeof(float)); 
   cudaMemcpy(device_matrices, matrices, nb_matrices * ALPH_SIZE * MATRIX_MAX_LENGTH * sizeof(float), cudaMemcpyHostToDevice);  

   // allocate device memory for lengths
   int *device_lengths; 
   cudaMalloc((void**) &device_lengths, nb_matrices * sizeof(int)); 
   cudaMemcpy(device_lengths, lengths, nb_matrices * sizeof(int), cudaMemcpyHostToDevice);  

   // allocate device memory for thresholds
   float *device_thresholds; 
   cudaMalloc((void**) &device_thresholds, nb_matrices * sizeof(int)); 
   cudaMemcpy(device_thresholds, thresholds, nb_matrices * sizeof(int), cudaMemcpyHostToDevice);  

   // allocate device memory for seq and copy them
   unsigned char *device_seq; 
   cudaMalloc((void**) &device_seq, seq_length + MATRIX_MAX_LENGTH); 
   cudaMemcpy(device_seq, seq, seq_length, cudaMemcpyHostToDevice);  

   // circular padding
   cudaMemcpy(device_seq + seq_length, seq, MATRIX_MAX_LENGTH, cudaMemcpyHostToDevice);  

   
   // allocate device memory for results (number of occurrences found in each thread)
   int *device_results_occ; 
   cudaMalloc((void**) &device_results_occ, NB_THREADS_SCAN * sizeof(int)); 
   int *results_occ= (int *) malloc(NB_THREADS_SCAN * sizeof(int)); 

   int *device_results_pos; 
   cudaMalloc((void**) &device_results_pos, NB_THREADS_SCAN * sizeof(int) * RES_PER_THREAD); 
   int *results_pos = (int *) malloc(NB_THREADS_SCAN * sizeof(int) * RES_PER_THREAD); 

   int *device_results_mat; 
   cudaMalloc((void**) &device_results_mat, NB_THREADS_SCAN * sizeof(int) * RES_PER_THREAD); 
   int *results_mat = (int *) malloc(NB_THREADS_SCAN * sizeof(int) * RES_PER_THREAD); 

   float *device_results_scores; 
#ifdef SCORES_OUT
   cudaMalloc((void**) &device_results_scores, NB_THREADS_SCAN * sizeof(float) * RES_PER_THREAD); 
   float *results_scores = (float *) malloc(NB_THREADS_SCAN * sizeof(float) * RES_PER_THREAD); 
#endif


   // allocate device memory for the slice tables
#ifdef SLICES
   float *device_tab = NULL;
   float *tab;
   tab = (float *) malloc(NBSLICES * SLICESIZE * NBMAT_PER_SLICE * sizeof(float));
   cudaMalloc((void**) &device_tab, NBSLICES * SLICESIZE * NBMAT_PER_SLICE * sizeof(float));
   float *pmatrices = matrices;
   int *plengths = lengths;
#endif

   // execute the kernel
   dim3  grid(NB_BLOCKS_SCAN, 1, 1);
   dim3  threads(NB_THREADS_PER_BLOCK_SCAN, 1, 1);
   printf("== Blocks: %4d   Threads/block: %4d\n", NB_BLOCKS_SCAN, NB_THREADS_PER_BLOCK_SCAN);

   int nb_matrices_treated = 0;
   int from = 0;
   int total_nbocc = 0;
   while (nb_matrices_treated < nb_matrices) {

      int total_length = 0;

#ifdef LATE_MATRIX_COPY
      int to = nb_matrices;
#else
      int to = from;
#ifdef SLICES  
      // ensure that the number of matrices in the slice data structure does not exceed the available memory
      while (to < nb_matrices && (to - from) < NBMAT_PER_SLICE) {
	 total_length += lengths[to];
	 to++;
      }
#else 
      // ensure that the sum of lengths of the matrices does not exceed the max total length
      while (to < nb_matrices && total_length + lengths[to] < MAX_TOTAL_LENGTH) {
	 total_length += lengths[to];
	 to++;
      }
#endif
#endif


      printf("== matrices %d to %d (%d matrices), total length %d\n", 
	     from, to-1, to - from, total_length);

#ifdef SLICES      
      device_group_multiple_matrices (tab, pmatrices, plengths, to - from);
      cudaMemcpy(device_tab, tab, NBSLICES * SLICESIZE * NBMAT_PER_SLICE * sizeof(float), cudaMemcpyHostToDevice);  
      // DEBUG
      DEBUG {
	 int s;
	 for (s = 0; s < NBSLICES; s++) {
	    printf("slice %2d : [ ",s);
	    int w;
	    for (w = 0; w < SLICESIZE * NBMAT_PER_SLICE; w++) {
	       printf("%.2f ",*(tab + s * SLICESIZE*NBMAT_PER_SLICE+w));
	    }
	    printf("]\n");
	 }
      }
#endif
            
      scan_pwm <<< grid, threads >>> (device_matrices, device_lengths, 
				      to - from, nb_matrices_treated,
				      device_thresholds, 
				      device_results_mat, device_results_pos, device_results_scores, device_results_occ,
				      device_seq, seq_length
#ifdef SLICES
				      , device_tab
#endif
				      );
      
      CUT_CHECK_ERROR("kernel 'scan' failed");

   
      cudaMemcpy(results_occ, device_results_occ, NB_THREADS_SCAN * sizeof(int), cudaMemcpyDeviceToHost);  
      cudaMemcpy(results_mat, device_results_mat, NB_THREADS_SCAN * sizeof(int) * RES_PER_THREAD, cudaMemcpyDeviceToHost);  
      cudaMemcpy(results_pos, device_results_pos, NB_THREADS_SCAN * sizeof(int) * RES_PER_THREAD, cudaMemcpyDeviceToHost);  
#ifdef SCORES_OUT
      cudaMemcpy(results_scores, device_results_scores, NB_THREADS_SCAN * sizeof(float) * RES_PER_THREAD, cudaMemcpyDeviceToHost);  
#endif
   
      display_results(results_occ, 
		      results_mat, results_pos, results_scores,
		      seq_length, to - from, display_hits);
      
      // prepare for the next iteration

      nb_matrices_treated += to - from;

#ifdef SLICES
      pmatrices += ALPH_SIZE * total_length;
      plengths += to - from;
#endif

      device_matrices += ALPH_SIZE * total_length;
      device_lengths += to-from;


      from = to;     

      for (int i=0; i<NB_THREADS_SCAN; i++) {
	 total_nbocc += results_occ[i] ;
      }

   }
   
   printf("Number of occurrences : %d\n",total_nbocc);

}








